/*
    Copyright (C) 1999 by  XCIN TEAM

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/      


#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "oximtool.h"
#include "imodule.h"

/*----------------------------------------------------------------------------

	Dynamic module loading:

	For ELF systems (Solaris, GNU/Linux, FreeBSD, HP-UX.11)
	For HP-UX PA32, PA64 systems.

----------------------------------------------------------------------------*/

#ifdef	HAVE_DLOPEN			/* ELF */
#include <dlfcn.h>
#else
#ifdef	HAVE_SHL_LOAD			/* HP-UX */
#include <dl.h>
#endif
#endif

#define BUFLEN	1024

struct _mod_stack_s {
   void *ld;
   mod_header_t *modp;
   int ref;
   struct _mod_stack_s *next;
};

struct _mod_stack_s *mod_stack = NULL;

/*----------------------------------------------------------------------------

	Load and Unload modules.

----------------------------------------------------------------------------*/

static int
find_module(char *path, int path_size, oxim_rc_t *xrc, char *sub_path)
{
    char fn[BUFLEN];

    if (check_datafile(path, sub_path, xrc, fn, BUFLEN) == False)
	return False;

    strcpy(path, fn);
    return check_file_exist(path, FTYPE_FILE);
}

mod_header_t *
load_module(char *modname, enum mtype mod_type, char *version,
	    oxim_rc_t *xrc, char *sub_path)
{
    struct _mod_stack_s *ms=mod_stack;
    char ldfn[BUFLEN];
    mod_header_t *module;
    void *ld=NULL;
    int err=1;

    while (ms != NULL) {
	if (strcmp(modname, ms->modp->name) == 0) {
	    ms->ref ++;
	    return ms->modp;
	}
	ms = ms->next;
    }

    snprintf(ldfn, BUFLEN, "%s.so", modname);
    if (find_module(ldfn, BUFLEN, xrc, sub_path) == True &&
#ifdef	HAVE_DLOPEN
        (ld = dlopen(ldfn, RTLD_LAZY)) != NULL)
#else
#ifdef	HAVE_SHL_LOAD
        (ld = (void *)shl_load(ldfn, BIND_DEFERRED, 0L)) != NULL)
#endif
#endif
	err = 0;

    if (err) {
#ifdef	HAVE_DLOPEN
	char *errstr = dlerror();
        perr(OXIMMSG_IWARNING, N_("dlerror: %s\n"), errstr);
#endif
	ld = NULL;
    }

#ifdef	HAVE_DLOPEN
    if (!err && !(module = (mod_header_t *)dlsym(ld, "module_ptr")))
#else
#ifdef	HAVE_SHL_LOAD
    if (!err && shl_findsym((shl_t *)&ld, "module_ptr", TYPE_DATA, &module)!=0)
#endif
#endif
    {
	perr(OXIMMSG_IWARNING, N_("module symbol \"module_ptr\" not found.\n"));
	err = 1;
    }
    if (!err && module->module_type != mod_type) {
	perr(OXIMMSG_IWARNING,
	     N_("invalid module type, type %d required.\n"), mod_type);
	err = 1;
    }
    if (!err && strcmp(module->version, version) != 0) {
	perr(OXIMMSG_IWARNING,
	     N_("invalid module version: %s, version %s required.\n"),
	     module->version, version);
    }

    if (err) {
        perr(OXIMMSG_WARNING,
	     N_("cannot load module \"%s\", ignore.\n"), modname);
	if (ld)
#ifdef HAVE_DLOPEN
	    dlclose(ld);
#else
#ifdef HAVE_SHL_OPEN
	    shl_unload((shl_t)ld);
#endif
#endif
	return NULL;
    }
    else {
	ms = oxim_malloc(sizeof(struct _mod_stack_s), 0);
	ms->ld = ld;
	ms->modp = module;
	ms->ref = 1;
	ms->next = mod_stack;
	mod_stack = ms;
	return module;
    }
}

void
unload_module(mod_header_t *modp)
{
    struct _mod_stack_s *ms = mod_stack;

    while (ms != NULL) {
	if (modp == ms->modp) {
	    ms->ref --;
	    break;
	}
	ms = ms->next;
    }
    if (ms && ms->ref <= 0) {
#ifdef	HAVE_DLOPEN
	dlclose(ms->ld);
#else
#ifdef	HAVE_SHL_LOAD
	shl_unload((shl_t)(ms->ld));
#endif
#endif
	mod_stack = ms->next;
	free(ms);
    }
}

/*----------------------------------------------------------------------------

	Print the module comment.

----------------------------------------------------------------------------*/

void
module_comment(mod_header_t *modp, char *mod_name)
{
    if (modp) {
	perr(OXIMMSG_NORMAL, N_("module \"%s\":"), mod_name);
	if (modp->comments)
	    perr(OXIMMSG_EMPTY, "\n\n%s\n", N_(modp->comments));
	else
	    perr(OXIMMSG_EMPTY, N_("no comments available.\n"));
    }
}

#include "imodule.h"

typedef struct {
    char *modname;
    char *objname;
    imodule_t *imodp;
} im_t;

static im_t IM[MAX_IM_ENTRY];

/*---------------------------------------------------------------------------

        Load Module & Module Init.

---------------------------------------------------------------------------*/

static int
check_module(module_t *modp, char *objenc)
{
    char *str=NULL, **objn, objname[64];
    int check_ok=0;

    if (modp == NULL)
	return False;
/*
 *  Check for necessary symbols.
 */
    if (! modp->conf_size)
	str = "conf_size";
    else if (! modp->init)
	str = "init";
    else if (! modp->xim_init)
	str = "xim_init";
    else if (! modp->xim_end)
	str = "xim_end";
    else if (! modp->keystroke)
	str = "keystroke";
    else if (! modp->show_keystroke)
	str = "show_keystroke";
    if (str) {
	perr(OXIMMSG_IWARNING,
	    N_("undefined symbol \"%s\" in module \"%s\", ignore.\n"),
	    str, modp->module_header.name);
	return  False;
    }
/*
 *  Check for the valid objname.
 */
    strncpy(objname, objenc, 64);
    if ((str = strrchr(objname, '@')))
	*str = '\0';
    objn = modp->valid_objname;
    if (! objn) {
	if (strcmp_wild(modp->module_header.name, objname) == 0)
	    check_ok = 1;
    }
    else {
	while (*objn) {
	    if(strcmp_wild(*objn, objname) == 0) {
		check_ok = 1;
		break;
	    }
	    objn ++;
	}
    }
    if (check_ok == 0) {
	perr(OXIMMSG_WARNING,
	    N_("invalid objname \"%s\" for module \"%s\", ignore.\n"),
	    objname, modp->module_header.name);
	return False;
    }
    return True;
}

static imodule_t *
creat_module(module_t *templet, char *objenc)
{
    imodule_t *imodp;

    imodp = oxim_malloc(sizeof(imodule_t), 1);
    imodp->modp = (void *)templet;
    imodp->name = templet->module_header.name;
    imodp->version = templet->module_header.version;
    imodp->comments = templet->module_header.comments;
    imodp->module_type = templet->module_header.module_type;
    imodp->conf = oxim_malloc(templet->conf_size, 1);
    imodp->init = templet->init;
    imodp->xim_init = templet->xim_init;
    imodp->xim_end = templet->xim_end;
    imodp->keystroke = templet->keystroke;
    imodp->show_keystroke = templet->show_keystroke;
    imodp->terminate = templet->terminate;

    imodp->objname = (objenc) ? (char *)strdup(objenc) : imodp->name; 

    return imodp;
}

static imodule_t *
IM_load(char *modname, char *objenc, oxim_rc_t *xrc)
{
    module_t *modp;
    imodule_t *imodp;
    int load_ok=1;

    modp = (module_t *)load_module(modname,MTYPE_IM,MODULE_VERSION,xrc, "modules");
    if (check_module(modp, objenc) == False)
	load_ok = 0;
    else {
	imodp = creat_module(modp, objenc);
	if (imodp->init(imodp->conf, objenc, xrc) != True) {
	    free(imodp->conf);
	    free(imodp);
	    load_ok = 0;
	}
    }
    if (load_ok == 0) {
	perr(OXIMMSG_WARNING,
	    N_("cannot load IM: %s, ignore.\n"), objenc);
	unload_module((mod_header_t *)modp);
	return NULL;
    }
    return imodp;
}

/*----------------------------------------------------------------------------

	Cinput structer manager.

----------------------------------------------------------------------------*/

int
IM_register(int idx, char *modname, char *objname, char *locale)
{
    im_t *imp;

    if (idx < 0 || idx >= MAX_IM_ENTRY)
	return False;

    imp = IM + idx;
    if (imp->modname)
	free(imp->modname);
    if (imp->objname)
	free(imp->objname);
    imp->modname = (char *)strdup(modname);
    imp->objname = (char *)strdup(objname);

    return True;
}

int
IM_check_registered(int idx)
{
    if (idx < 0 || idx >= MAX_IM_ENTRY)
	return False;
    if (IM[idx].modname && IM[idx].objname)
	return True;
    else
	return False;
}

imodule_t *
IM_get(int idx, oxim_rc_t *xrc)
{
    im_t *imp = IM + idx;

    if (idx < 0 || idx >= MAX_IM_ENTRY)
	return NULL;
    if (imp->modname && imp->objname && imp->imodp == NULL)
	imp->imodp = IM_load(imp->modname, imp->objname, xrc);
    if (imp->imodp == NULL)
	IM_free(idx);
    return imp->imodp;
}

imodule_t *
IM_get_next(int idx, int *idx_ret, oxim_rc_t *xrc)
{
    int i, j;
    im_t *imp;

    if (idx < 0 || idx >= MAX_IM_ENTRY)
	return NULL;

    *idx_ret = -1;
    for (j=0, i=idx, imp=IM+idx; j<MAX_IM_ENTRY; j++, i++, imp++) {
	if (i >= MAX_IM_ENTRY) {
	    imp = IM;
	    i = 0;
	}
	if (imp->modname && imp->objname)
	{
	    if (imp->imodp == NULL)
	    {
		imp->imodp = IM_load(imp->modname, imp->objname, xrc);
		if (imp->imodp != NULL )
		{
		    *idx_ret = i;
		    break;
		}
	    }
	}
    }
    if (imp->imodp == NULL)
	IM_free(*idx_ret);
    return imp->imodp;
}

imodule_t *
IM_get_prev(int idx, int *idx_ret, oxim_rc_t *xrc)
{
    int i, j;
    im_t *imp;

    if (idx < 0 || idx >= MAX_IM_ENTRY)
	return NULL;

    *idx_ret = -1;
    for (j=0, i=idx, imp=IM+idx; j<MAX_IM_ENTRY; j++, i--, imp--) {
	if (i < 0) {
	    imp = IM + MAX_IM_ENTRY - 1;
	    i = MAX_IM_ENTRY - 1;
	}
	if (imp->modname && imp->objname) {
	    *idx_ret = i;
	    break;
	}
    }
    if (*idx_ret != -1 && imp->modname && imp->objname && imp->imodp == NULL)
	imp->imodp = IM_load(imp->modname, imp->objname, xrc);
    if (imp->imodp == NULL)
	IM_free(*idx_ret);
    return imp->imodp;
}

numlist_t *
IM_get_numlist(void)
{
    static numlist_t numlist[MAX_IM_ENTRY+1] = {(numlist_t)-1};
    numlist_t *s=numlist;
    int i;

    if (numlist[0] != (numlist_t)-1)
	return numlist;

    for (i=0; i<MAX_IM_ENTRY; i++) {
	if (IM[i].modname && IM[i].objname) {
	    *s = (numlist_t)i;
	    s++;
	}
    }
    *s = (numlist_t)-1;

    return numlist;
}

imodule_t *
IM_search(char *objname, char *locale, int *idx_ret, oxim_rc_t *xrc)
{
    int i;
    im_t *imp;
    char objloc[100];

    strcpy(objloc, objname);

    *idx_ret = -1;
    for (i=0, imp=IM; i<MAX_IM_ENTRY; i++, imp++) {
	if (imp->objname && ! strcmp(imp->objname, objloc)) {
	    *idx_ret = i;
	    break;
	}
    }
    if (*idx_ret != -1 && imp->modname && imp->objname && imp->imodp == NULL)
	imp->imodp = IM_load(imp->modname, imp->objname, xrc);
    if (imp->imodp == NULL)
	IM_free(*idx_ret);
    return imp->imodp;
}

void 
IM_free(int idx)
{
    im_t *imp;

    if (idx < 0 || idx >= MAX_IM_ENTRY)
	return;

    imp = IM+idx;
    if (imp->modname) {
	free(imp->modname);
	imp->modname = NULL;
    }
    if (imp->objname) {
	free(imp->objname);
	imp->objname = NULL;
    }
    if (imp->imodp) {
	if (imp->imodp->terminate)
	    imp->imodp->terminate(imp->imodp->conf);
	if (imp->imodp->modp)
	    unload_module((mod_header_t *)imp->imodp->modp);
	if (imp->imodp->conf)
	    free(imp->imodp->conf);
	free(imp->imodp);
	imp->imodp = NULL;
    }
}

void
IM_free_all(void)
{
    int i;
    for (i=0; i<MAX_IM_ENTRY; i++)
	IM_free(i);
}

int
get_objenc(char *objname, objenc_t *objenc)
{
    int i;
    im_t *imp;
    char *s;

    for (i=0, imp=IM; i<MAX_IM_ENTRY; i++, imp++) {
	if (imp->objname && ! strcmp(imp->objname, objname))
	    break;
    }
    if (i >= MAX_IM_ENTRY)
	return -1;

    strcpy(objenc->objname, objname);
    strcpy(objenc->objloadname, objname);
    return 0;
}
