/*
    Copyright (C) 1999 by  XCIN TEAM

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>
#include "gui.h"
#include "oxim.h"

winlist_t *gui_root_init(gui_t *gui, xccore_t *xccore);
winlist_t *gui_status_init(gui_t *gui, xccore_t *xccore);
winlist_t *gui_preedit_init(gui_t *gui, xccore_t *xccore);
winlist_t *gui_select_init(gui_t *gui, xccore_t *xccore);
winlist_t *gui_tray_init(gui_t *gui, xccore_t *xccore);
int check_tray_manager(gui_t *gui);
void xim_terminate(void);
int ic_destroy(int icid, xccore_t *xccore);

static gui_t *gui;

void Xft_Init(xccore_t *xccore);
XftFont *gui_find_font(winlist_t *win, FcChar32 ucs4);
/*----------------------------------------------------------------------------

	Basic WinList handling functions.

----------------------------------------------------------------------------*/

static winlist_t *winlist, *free_win;

static void 
free_win_content(winlist_t *win)
{
    int i;

    if ((win->winmode & WMODE_EXIT))
	return;

    if (win->win_destroy_func)
	win->win_destroy_func(gui, win);
    else {
	XUnmapWindow(gui->display, win->window);
	win->winmode &= ~WMODE_MAP;

	XDestroyWindow(gui->display, win->window);
    }
    win->winmode |= WMODE_EXIT;
    /* Add by Firefly(firefly@ossii.com.tw) */
    XftDrawDestroy(win->draw);
}

winlist_t *
gui_search_win(Window window)
{
    winlist_t *w = winlist;

    while (w) {
	if (w->window == window)
	    return w;
	w = w->next;
    }
    return NULL;
}

winlist_t *
gui_new_win(void)
{
    winlist_t *win;

    if (free_win) {
	win = free_win;
	free_win = free_win->next;
	bzero(win, sizeof(winlist_t));
    }
    else
	win = oxim_malloc(sizeof(winlist_t), 1);
    win->next = winlist;
    winlist = win;

    return win;
}

void
gui_freewin(Window window)
{
    winlist_t *w = winlist, *wp = NULL;

    while (w) {
	if (w->window == window) {
	    free_win_content(w);
	    if (wp == NULL)
		winlist = w->next;
	    else
		wp->next = w->next;
	    w->next = free_win;
	    free_win = w;
	    return;
	}
	wp = w;
	w  = w->next;
    }
}

void
gui_winmap_change(winlist_t *win, int state)
{
    if ((win->winmode & WMODE_EXIT))
	return;

    if (state && !(win->winmode & WMODE_MAP))
    {
	XMapRaised(gui->display, win->window);
	win->winmode |= WMODE_MAP;
    }
    else if (!state && (win->winmode & WMODE_MAP))
    {
	XUnmapWindow(gui->display, win->window);
	win->winmode &= ~WMODE_MAP;
    }
}

int
gui_check_window(Window window)
{
    Window root;
    int x, y;
    unsigned width, height, bw, depth;

    XGetGeometry(gui->display, window,
		&root, &x, &y, &width, &height, &bw, &depth);
    XSync(gui->display, False);
    if (errstatus == 0)
	return True;
    else {
	errstatus = 0;
	return False;
    }
}

int
gui_check_input_focus(xccore_t *xccore, Window win)
{
    Window w;
    int revert_to_return;

    XGetInputFocus(xccore->gui.display, &w, &revert_to_return);
    return (w == win) ? True : False;
}

/*----------------------------------------------------------------------------

	GUI Monitoring the XIM client windows. 

----------------------------------------------------------------------------*/
/*
struct mw_s {
    Window win;
    int icid;
    xmode_t flag;
};

#define MW_LENTH	64
static struct mw_s *mw;
static int mw_edx, mw_len;

void
gui_set_monitor(Window w, xmode_t flag, int icid)
{
    int i, idx=-1;

    for (i=0; i<mw_edx; i++) {
	if (idx == -1 && mw[i].win == (Window)0)
	    idx = i;
	if (mw[i].win == w) {
	    mw[i].flag |= flag;
	    return;
	}
    }
    if (idx == -1) {
	if (mw_edx >= mw_len) {
	    mw_len += MW_LENTH;
	    mw = oxim_realloc(mw, sizeof(struct mw_s)*mw_len);
	    for (i=mw_edx; i<mw_len; i++)
		mw[i].win = (Window)0;
	}
	idx = mw_edx;
	mw_edx ++;
    }
    XSelectInput(gui->display, w, StructureNotifyMask);
    mw[idx].win = w;
    mw[idx].icid = icid;
    mw[idx].flag = flag;
}

static void
gui_unset_monitor(Window w, xccore_t *xccore)
{
    int i;

    for (i=0; i<mw_edx; i++) {
	if (mw[i].win == w &&
	    (mw[i].flag & WIN_MONITOR_CLIENT)) {
	    ic_destroy(mw[i].icid, xccore);
	    mw[i].win = (Window)0;
	    mw[i].icid = 0;
	    mw[i].flag = 0;
	    if (i == mw_edx-1 && mw_edx > 0)
		mw_edx --;
	    break;
	}
    }
}

static void
gui_check_monitor(XConfigureEvent *event)
{
    int i;

    for (i=0; i<mw_edx; i++) {
	if (mw[i].win == event->window &&
	    (mw[i].flag & WIN_MONITOR_OVERSPOT)) {
	    gui_overspot_check_client(gui, mw[i].icid);
	    break;
	}
    }
}
*/
/*----------------------------------------------------------------------------

	GUI Initialization & Main loop

----------------------------------------------------------------------------*/

static int errhandle(Display *disp, XErrorEvent *xevent)
{
    errstatus = xevent->error_code;
    return 1;
}

void
gui_init(xccore_t *xccore)
{
    gui = &(xccore->gui);

    if (! (gui->display = XOpenDisplay(xccore->irc->display_name)))
	perr(OXIMMSG_ERROR, N_("cannot open display: %s\n"), 
			    xccore->irc->display_name);
    (void) XSetErrorHandler(errhandle);
    gui->screen = DefaultScreen(gui->display);
    gui->visual = DefaultVisual(gui->display, gui->screen);
    gui->colormap = DefaultColormap(gui->display, gui->screen);
    gui->display_width = DisplayWidth(gui->display, gui->screen);
    gui->display_height = DisplayHeight(gui->display, gui->screen);
    gui->root = RootWindow(gui->display, gui->screen);
    gui->wm_del_win  = XInternAtom(gui->display, "WM_DELETE_WINDOW", False);

    /* 指定初值 */
    unsigned int i;
    //for (i=0 ; i < MAX_COLOR; i++)
//	gui->colors[i] = NULL;

    oxim_rc_t *xrc = &(xccore->oxim_rc);
    char *cmd[1], value[256];

    cmd[0] = "WINBOX_COLOR";
    if (get_resource(xrc, cmd, value, 256, 1))
	XftColorAllocName(gui->display, gui->visual, gui->colormap,
		value, &gui->colors[WINBOX_COLOR]);

    cmd[0] = "BORDER_COLOR";
    if (get_resource(xrc, cmd, value, 256, 1))
	XftColorAllocName(gui->display, gui->visual, gui->colormap,
		value, &gui->colors[BORDER_COLOR]);

    cmd[0] = "LIGHT_COLOR";
    if (get_resource(xrc, cmd, value, 256, 1))
	XftColorAllocName(gui->display, gui->visual, gui->colormap,
		value, &gui->colors[LIGHT_COLOR]);

    cmd[0] = "DARK_COLOR";
    if (get_resource(xrc, cmd, value, 256, 1))
	XftColorAllocName(gui->display, gui->visual, gui->colormap,
		value, &gui->colors[DARK_COLOR]);

    cmd[0] = "CURSOR_COLOR";
    if (get_resource(xrc, cmd, value, 256, 1))
	XftColorAllocName(gui->display, gui->visual, gui->colormap,
		value, &gui->colors[CURSOR_COLOR]);

    cmd[0] = "CURSORFONT_COLOR";
    if (get_resource(xrc, cmd, value, 256, 1))
	XftColorAllocName(gui->display, gui->visual, gui->colormap,
		value, &gui->colors[CURSORFONT_COLOR]);

    cmd[0] = "FONT_COLOR";
    if (get_resource(xrc, cmd, value, 256, 1))
	XftColorAllocName(gui->display, gui->visual, gui->colormap,
		value, &gui->colors[FONT_COLOR]);

    cmd[0] = "CONVERTNAME_COLOR";
    if (get_resource(xrc, cmd, value, 256, 1))
	XftColorAllocName(gui->display, gui->visual, gui->colormap,
		value, &gui->colors[CONVERTNAME_COLOR]);

    cmd[0] = "INPUTNAME_COLOR";
    if (get_resource(xrc, cmd, value, 256, 1))
	XftColorAllocName(gui->display, gui->visual, gui->colormap,
		value, &gui->colors[INPUTNAME_COLOR]);

    cmd[0] = "UNDERLINE_COLOR";
    if (get_resource(xrc, cmd, value, 256, 1))
	XftColorAllocName(gui->display, gui->visual, gui->colormap,
		value, &gui->colors[UNDERLINE_COLOR]);

    cmd[0] = "KEYSTROKE_COLOR";
    if (get_resource(xrc, cmd, value, 256, 1))
	XftColorAllocName(gui->display, gui->visual, gui->colormap,
		value, &gui->colors[KEYSTROKE_COLOR]);

    cmd[0] = "KEYSTROKEBOX_COLOR";
    if (get_resource(xrc, cmd, value, 256, 1))
	XftColorAllocName(gui->display, gui->visual, gui->colormap,
		value, &gui->colors[KEYSTROKEBOX_COLOR]);

    cmd[0] = "SELECTFONT_COLOR";
    if (get_resource(xrc, cmd, value, 256, 1))
	XftColorAllocName(gui->display, gui->visual, gui->colormap,
		value, &gui->colors[SELECTFONT_COLOR]);

    cmd[0] = "SELECTBOX_COLOR";
    if (get_resource(xrc, cmd, value, 256, 1))
	XftColorAllocName(gui->display, gui->visual, gui->colormap,
		value, &gui->colors[SELECTBOX_COLOR]);

    Xft_Init(xccore);

    gui->root_win    = gui_root_init(gui, xccore);
    gui->status_win  = gui_status_init(gui, xccore);
    gui->preedit_win = gui_preedit_init(gui, xccore);
    gui->select_win  = gui_select_init(gui, xccore);
    gui->tray_win    = gui_tray_init(gui, xccore);
    gui->winchange |= WIN_CHANGE_IM;

}

void 
gui_loop(xccore_t *xccore)
{
    XEvent event;
    winlist_t *win;

    while (1) {
	if ((xccore->oxim_mode & OXIM_RUN_EXITALL)) {
	    xim_terminate();
	    IM_free_all();
	    exit(0);
	}
	check_tray_manager(gui);
        XNextEvent(gui->display, &event);
	if (XFilterEvent(&event, None) == True)
            continue;

	switch (event.type) {
	case Expose:
	    if (event.xexpose.count == 0) {
		win = gui_search_win(event.xexpose.window);
		if (win && win->win_draw_func)
		    win->win_draw_func(gui, win);
	    }
	    break;
	/* Add by Firefly(firefly@ossii.com.tw) */
	/* 按下滑鼠左鍵，表示要拖曳視窗 */
	case ButtonPress:
	    if (event.xbutton.button == Button1)
	    {
		win = gui_search_win(event.xexpose.window);
		if (win && win->win_mouse_func)
		    win->win_mouse_func(gui, win, &(event));
	    }
	    break;
	/*--------------------------------------*/
	case GraphicsExpose:
	    win = winlist;
	    while (win) {
		if (win->win_draw_func)
		    win->win_draw_func(gui, win);
		win = win->next;
	    }
	    break;
	case ConfigureNotify:
/*
	    win = gui_search_win(event.xconfigure.window);
	    if (! win)
		gui_check_monitor(&(event.xconfigure));
	    else if (win->win_attrib_func)
	    if (win->win_attrib_func)
		win->win_attrib_func(gui, win, &(event.xconfigure),
				(xccore->oxim_mode & OXIM_KEEP_POSITION));
*/
	    break;
/*
	case DestroyNotify:
	    gui_unset_monitor(event.xdestroywindow.window, xccore);
	    break;
*/
	case ClientMessage:
	    if (event.xclient.format == 32 && 
		event.xclient.data.l[0] == gui->wm_del_win)
		gui_freewin(event.xclient.window);
	    break;
        }
    }
}

/*----------------------------------------------------------------------------

	GUI update window list functions.

----------------------------------------------------------------------------*/

Window
gui_request_init(IM_Context_t *imc, greq_win_t *gw)
{
    Window window;

    switch (gw->greq_data->type) {
	window = (Window)0;
	break;
    }
    return window;
}

static void
gui_bell(xmode_t oxim_mode)
{
    static int bell_pitch;

    if (! (oxim_mode & OXIM_DIFF_BEEP))
	XBell(gui->display, -20);
    else {
	XKeyboardControl kc;
	XKeyboardState ks;
	int new_pitch;

	if (bell_pitch == 0) {
	    XGetKeyboardControl(gui->display, &ks);
	    bell_pitch = ks.bell_pitch;
	}
	if ((gui->winchange & WIN_CHANGE_BELL))
	    new_pitch = 700;
	else if ((gui->winchange & WIN_CHANGE_BELL2))
	    new_pitch = 1400;
	else
	    new_pitch = bell_pitch;

	kc.bell_pitch = new_pitch;
	XChangeKeyboardControl(gui->display, KBBellPitch, &kc);
	XBell(gui->display, -20);
	kc.bell_pitch = bell_pitch;
	XChangeKeyboardControl(gui->display, KBBellPitch, &kc);
    }
}

void
gui_update_winlist(xccore_t *xccore)
{
    gui->preedit_win->win_draw_func(gui, gui->preedit_win);

    if ((gui->winchange & WIN_CHANGE_IM)) {
	gui->status_win->win_draw_func(gui, gui->status_win);
    }

    gui->tray_win->win_draw_func(gui, gui->tray_win);

    if ((gui->winchange & WIN_CHANGE_BELLALL))
	gui_bell(xccore->oxim_mode);
    gui->winchange = 0;
}

/*----------------------------------------------------------------------------
    Mouse Event
----------------------------------------------------------------------------*/
static void gui_get_mouse_xy(gui_t *gui, int *x, int *y)
{
    Window ret_root, ret_child;
    int win_x, win_y, mx, my;
    unsigned int mask;

    XQueryPointer(gui->display, gui->root, &ret_root, &ret_child, &win_x, &win_y, &mx, &my, &mask);
    *x = mx;
    *y = my;
}

static void gui_draw_xor_box(gui_t *gui, GC gc, int x, int y, int width, int height)
{
    XSetForeground(gui->display, gc, WhitePixel(gui->display, gui->screen) ^ BlackPixel(gui->display, gui->screen));
    XDrawRectangle(gui->display, gui->root, gc, x, y, width, height);
}

void
gui_mouse_event(gui_t *gui, winlist_t *win, XEvent *event)
{
    int event_x, event_y;
    int offset_x, offset_y;
    int move_x, move_y;
    GC  moveGC;
    int draw_flag = False;

    gui_get_mouse_xy(gui, &event_x, &event_y);
    offset_x = event_x - win->pos_x;
    offset_y = event_y - win->pos_y;

    moveGC = XCreateGC(gui->display, gui->root, 0, NULL);
    XSetSubwindowMode(gui->display, moveGC, IncludeInferiors);
    XSetForeground(gui->display, moveGC, BlackPixel(gui->display, gui->screen));
    XSetFunction(gui->display, moveGC, GXxor);

    XChangeActivePointerGrab(gui->display,
                PointerMotionMask | ButtonMotionMask | ButtonReleaseMask |
                OwnerGrabButtonMask, None, CurrentTime);

    XGrabServer(gui->display);

    XEvent myevent;
    while(1)
    {
	XNextEvent(gui->display, &myevent);
	switch(myevent.type)
	{
	    case ButtonRelease:
		if(myevent.xbutton.button == Button1)
		{
		    if (draw_flag)
			gui_draw_xor_box(gui, moveGC, move_x - offset_x, move_y - offset_y, win->width, win->height);
		    XFreeGC(gui->display, moveGC);
		    gui_get_mouse_xy(gui, &move_x, &move_y);
		    win->pos_x = move_x - offset_x;
		    win->pos_y = move_y - offset_y;

		    /* */
		    if (win->pos_x < 0)
			win->pos_x = 0;
		    if (win->pos_y < 0)
			win->pos_y = 0;
		    if (win->pos_x + win->width > gui->display_width)
			win->pos_x = gui->display_width - win->width;
		    if (win->pos_y + win->height > gui->display_height)
			win->pos_y = gui->display_height - win->height;

		    XMoveWindow(gui->display, win->window,
				 win->pos_x, win->pos_y);
		    XUngrabServer(gui->display);
		    return;
		}
		break;

	    case MotionNotify:
		if (draw_flag)
		    gui_draw_xor_box(gui, moveGC, move_x - offset_x, move_y - offset_y, win->width, win->height);
		gui_get_mouse_xy(gui, &move_x, &move_y);
		gui_draw_xor_box(gui, moveGC, move_x - offset_x, move_y - offset_y, win->width, win->height);
		draw_flag = True;
		break;
	    default:
		break;
	}
    }
}

/******************************************************************
*
* Add by Firefly(firefly@ossii.com.tw)
*
******************************************************************/
#include <X11/Xft/Xft.h>

/* 預設的字型名稱 */
static FcFontSet *default_fontset = NULL;

static unsigned int num_fonts = 0;
static XftFont **xftfonts = NULL;

static unsigned int default_fontsize = 0;
static unsigned int leading = 1;

void Xft_Init(xccore_t *xccore)
{
    default_fontsize = xccore->irc->default_font_size ? xccore->irc->default_font_size : 16;

    XftFontSet *fonts;
    char *familyName;
    int weight, slant, scalable;
    FcCharSet *charset;
    int i;

    default_fontset = FcFontSetCreate();

    fonts = XftListFonts(gui->display, gui->screen, (char *)0,
			XFT_FILE, XFT_INDEX,
			XFT_FAMILY, XFT_WEIGHT, XFT_SLANT, XFT_SPACING,
			FC_CHARSET, FC_SCALABLE,
			(char *)0);

    for (i = 0; i < fonts->nfont; i++)
    {
	FcPattern *pat = fonts->fonts[i];
	if (XftPatternGetString(pat, XFT_FAMILY, 0, &familyName) != XftResultMatch ||
	    XftPatternGetInteger(pat, XFT_WEIGHT, 0, &weight)!= XftResultMatch ||
	    XftPatternGetInteger(pat, XFT_SLANT, 0, &slant)!= XftResultMatch ||
	    XftPatternGetBool(pat, XFT_SCALABLE, 0, &scalable)!= XftResultMatch ||
	    FcPatternGetCharSet(pat, FC_CHARSET, 0, &charset)!= XftResultMatch)
	{
	    continue;
	}

	if (weight > FC_WEIGHT_MEDIUM || slant != FC_SLANT_ROMAN || !scalable)
	{
	    continue;
	}

	FcPattern *dup_pat = FcPatternDuplicate(pat);
	if (dup_pat)
	    FcFontSetAdd(default_fontset, dup_pat);
    }
    XftFontSetDestroy (fonts);

    num_fonts = 1;
    xftfonts = (XftFont **)oxim_malloc(sizeof(XftFont *), True);
    xftfonts[0] = XftFontOpen(gui->display, gui->screen,
		XFT_FAMILY, XftTypeString, xccore->irc->default_font_name,
		XFT_PIXEL_SIZE, XftTypeDouble, (double)default_fontsize,
		NULL);
}

/* 計算字串繪圖寬度 */
unsigned int Utf8_Char_Escapement(XftDraw *draw, XftFont *font, FcChar8 *Utf8Char, int Utf8len)
{
    XGlyphInfo GlyphInfo;
    XftTextExtentsUtf8(XftDrawDisplay(draw), font, Utf8Char, Utf8len,  &GlyphInfo);
    return GlyphInfo.xOff;
}

/* 繪出Utf8字元 */
unsigned int Draw_Utf8_Char(XftDraw *draw, XftFont *font, int foreground_idx, int background_idx, int x, int y, FcChar8 *Utf8Char, int Utf8len)
{
    unsigned int width = Utf8_Char_Escapement(draw, font, Utf8Char, Utf8len);
    if (background_idx >= 0 && background_idx <= MAX_COLOR)
    {
	XftDrawRect(draw, &gui->colors[background_idx], x, y - default_fontsize+1, width, default_fontsize);
    }

    XftDrawStringUtf8(draw, &gui->colors[foreground_idx], font, x, y, Utf8Char, Utf8len);
    return width;
}

XftFont *gui_find_font(winlist_t *win, FcChar32 ucs4)
{
    unsigned int i;

    /* */
    for (i=0 ; i < num_fonts ; i++)
    {
	if (XftCharExists(gui->display, xftfonts[i], ucs4))
	    return xftfonts[i];
    }

    XftFont *font = NULL;
    for (i=0; i< default_fontset->nfont; i++)
    {
	FcPattern *pat = default_fontset->fonts[i];
	FcCharSet *fcs = NULL;

	if (FcPatternGetCharSet(pat, FC_CHARSET, 0, &fcs) != XftResultMatch)
	    continue;

	if (!FcCharSetHasChar(fcs, ucs4))
	    continue;

	XftResult res;
	XftPattern *mpat = XftFontMatch(gui->display, gui->screen, pat, &res);
	if (!mpat)
	    continue;

	FcPattern *dup = XftPatternDuplicate(mpat);
	XftPatternAddInteger(dup, XFT_PIXEL_SIZE, gui_FontSize(win));
	XftFont *chkfont = XftFontOpenPattern(gui->display, dup);

        if (!chkfont)
	{
	    XftPatternDestroy(dup);
	}
	else
	{
	    xftfonts = (XftFont **)realloc(xftfonts, (num_fonts +1) * sizeof(XftFont *));
	    xftfonts[num_fonts++] = chkfont;
	    return chkfont;
	}
    }

    return xftfonts[0];
}

unsigned int gui_Draw_String(winlist_t *win, int foreground_idx, int background_idx, int x, int y, char *string, int len)
{
    char *p = string;
    FcChar32 ucs4;
    unsigned int string_width = 0;
    int nbytes;

    while (len && (nbytes = FcUtf8ToUcs4((FcChar8 *)p, &ucs4, len)) > 0)
    {
        unsigned int width;

	XftFont *font = gui_find_font(win, ucs4);

	width = Draw_Utf8_Char(win->draw, font, foreground_idx, background_idx, x, y, (FcChar8 *)p, nbytes);

	x += width;
	string_width += width;
	p += nbytes;
	len -= nbytes;
    }
    return string_width;
}

unsigned int gui_TextEscapement(winlist_t *win, char *string, int len)
{
    char *p = string;
    FcChar32 ucs4;
    unsigned int string_width = 0;
    int nbytes = utf8_to_ucs4(p, &ucs4, len);

    while (len && (nbytes = utf8_to_ucs4(p, &ucs4, len)) > 0)
    {
	XftFont *font = gui_find_font(win, ucs4);

	unsigned int width = Utf8_Char_Escapement(win->draw, font, (FcChar8 *)p, nbytes);

	string_width += width;
	p += nbytes;
	len -= nbytes;
    }

    return string_width;
}

void gui_Draw_3D_Box(winlist_t *win, int x, int y, int width, int height, int fill_idx, int boxstyle)
{
    GC lightGC = XCreateGC(gui->display, win->window, 0, NULL);
    GC darkGC  = XCreateGC(gui->display, win->window, 0, NULL);

    if (boxstyle)
    {
	XSetForeground(gui->display, lightGC, gui_Color(LIGHT_COLOR));
	XSetForeground(gui->display, darkGC, gui_Color(DARK_COLOR));
    }
    else
    {
	XSetForeground(gui->display, lightGC, gui_Color(DARK_COLOR));
	XSetForeground(gui->display, darkGC, gui_Color(LIGHT_COLOR));
    }
    XSetLineAttributes(gui->display, lightGC, 1, LineSolid, CapRound, JoinRound);
    XSetLineAttributes(gui->display, darkGC, 1, LineSolid, CapRound, JoinRound);
    
    int x2 = x + width - 1;
    int y2 = y + height - 1;
    XDrawLine(gui->display, win->window, lightGC, x, y, x2, y);
    XDrawLine(gui->display, win->window, lightGC, x, y, x, y2);
    XDrawLine(gui->display, win->window, darkGC, x, y2, x2, y2);
    XDrawLine(gui->display, win->window, darkGC, x2, y, x2, y2);
    if (fill_idx >= 0 && fill_idx <= MAX_COLOR)
    {
	XftDrawRect(win->draw, &gui->colors[fill_idx], x+1, y + 1, width-2, height-2);
    }
    XFreeGC(gui->display, lightGC);
    XFreeGC(gui->display, darkGC);
}

unsigned long gui_Color(int color_index)
{
    return gui->colors[color_index].pixel;
}

unsigned int gui_FontSize(winlist_t *win)
{

    if (win->font_size)
	return win->font_size;

    return default_fontsize;
}

void gui_Draw_Image(winlist_t *win, int x, int y, char **xpm_data)
{
    XImage* image;

    XpmCreateImageFromData(gui->display, xpm_data, &image, NULL, NULL);
    XPutImage(gui->display, win->window, DefaultGC(gui->display, gui->screen),
                image, 0, 0, x, y, image->width, image->height);
    XDestroyImage(image);
}
