#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/xpm.h>
#include <string.h>
#include "oximtool.h"
#include "gui.h"
#include "oxim.h"

#include "images/halfchar.xpm"
#include "images/fullchar.xpm"
#include "images/out_cn.xpm"
#include "images/out_tw.xpm"
#include "images/sideback_l.xpm"
#include "images/sideback.xpm"
#include "images/sideback_r.xpm"

#define X_LEADING       2
#define Y_LEADING       2

static void 
win_draw(gui_t *gui, winlist_t *win, IC *ic)
{
    char *inpn=NULL;
    char **inpb=NULL, **inp_conv=NULL;
    char *s, buf[100];
    IM_Context_t *imc = ic->imc;
    int x, y, len, img_y;
    img_y = (win->height - 16) / 2;
    inpb = ((imc->inp_state & IM_2BYTES)) ? icon_full : icon_half;

    if (imc->inp_state & IM_OUTSIMP)
	inp_conv = icon_cn;
    else if (imc->inp_state & IM_OUTTRAD)
	inp_conv = icon_tw;

    if ((imc->inp_state & IM_CINPUT)) {
        s = imc->inpinfo.inp_cname;
        while (*s) {
            if (*s == '%' && *(s+1) == '%' && *(s+2)) {
                inpn = s+2;
                break;
            }
            s++;
        }
        if (! inpn) {
            strcpy((char *)buf, imc->inpinfo.inp_cname);
            inpn = buf;
        }
    }
    else
        inpn = gui->inpn_english;

    len = strlen((char *)inpn);

    gui_winmap_change(win, 1);
    XClearWindow(gui->display, win->window);
    
    gui_Draw_Image(win, 0, 0, sideback_l);
    x = 5 ;
    y = win->height - ((win->height - win->font_size) / 2) - 1;

    int inpn_w = gui_TextEscapement(win, inpn, len);

    gui_Draw_String(win, FONT_COLOR, NO_COLOR, x+1, y+1, inpn, len);
    x += gui_Draw_String(win, LIGHT_COLOR, NO_COLOR, x, y, inpn, len);

    x += (X_LEADING * 2) + 1;
    gui_Draw_Image(win, x, img_y, inpb);
    x += 16 + (X_LEADING * 2);

    if (inp_conv)
    {
	gui_Draw_Image(win, x, img_y, inp_conv);
	x += 20;
    }

    win->width = x + 5;
    gui_Draw_Image(win, win->width - 5, 0, sideback_r);

    if (win->width + win->pos_x > gui->display_width)
    {
	win->pos_x = gui->display_width - win->width;
	XMoveWindow(gui->display, win->window, win->pos_x, win->pos_y);
    }
    XResizeWindow(gui->display, win->window, win->width, win->height);
    XFlush(gui->display);
}

static void
gui_status_draw(gui_t *gui, winlist_t *win)
{
    xccore_t *xccore = (xccore_t *)win->data;
    IC *ic = xccore->ic;
    inp_state_t inp_state = (ic) ? ic->imc->inp_state : 0;
    if (((win->winmode & WMODE_EXIT)) || !ic || (!(inp_state & IM_CINPUT) && !(inp_state & IM_2BYTES))) {
	gui_winmap_change(win, 0);
	return;
    }
    win_draw(gui, win, ic);
}

winlist_t *
gui_status_init(gui_t *gui, xccore_t *xccore)
{
    unsigned int fontsize = xccore->irc->status_font_size ? xccore->irc->status_font_size : 16;
    winlist_t *win = gui_new_win();
    win->width    = 1;
    win->height   = 24;
/*  Winlist Setup  */
    win->pos_x = gui->display_width - 200;
    win->pos_y = gui->display_height - 64;
    win->window = XCreateSimpleWindow(gui->display, gui->root, 
		win->pos_x, win->pos_y, win->width, win->height, 0, 0, 0);
    win->draw = XftDrawCreate(gui->display, win->window,
                DefaultVisual(gui->display, gui->screen),
                gui->colormap);

    if (fontsize < 12 || fontsize > 18)
	win->font_size = 16;
    else
	win->font_size = fontsize;

    win->data = (void *)xccore;
    win->win_draw_func = gui_status_draw;
    win->win_mouse_func = gui_mouse_event;
    win->win_attrib_func = NULL;
    win->win_destroy_func = NULL;

    XSetWindowAttributes win_attr;
    win_attr.override_redirect = True;
    win_attr.save_under = True;
    XChangeWindowAttributes(gui->display, win->window,
				CWOverrideRedirect|CWSaveUnder, &win_attr);
    XSelectInput(gui->display, win->window, (ExposureMask|ButtonPressMask|ButtonReleaseMask|StructureNotifyMask));

    Pixmap pix;
    XpmCreatePixmapFromData(gui->display, win->window, sideback, &pix, NULL, NULL);
    XSetWindowBackgroundPixmap(gui->display, win->window, pix);
    XFreePixmap(gui->display, pix);

    return win;
}

