/*
    Copyright (C) 1999 by  XCIN TEAM

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/      

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#ifdef HPUX
#  define _INCLUDE_POSIX_SOURCE
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <X11/Xlocale.h>
#include <signal.h>
#include <zlib.h>
#include "oximtool.h"
#include "oxim.h"
#include "fkey.h"
#include "../util/oxim2tab/syscin.h"

int verbose, errstatus;

static xccore_t oxim_core;		/* OXIM kernel data. */

void gui_init(xccore_t *xccore);
void gui_loop(xccore_t *xccore);
void xim_init(xccore_t *xccore);
void xim_add_locale(char *locale);

/*----------------------------------------------------------------------------

        Initial Input.

----------------------------------------------------------------------------*/

static void
oxim_setlocale(void)
{
    locale_t *locale = &(oxim_core.oxim_rc.locale);
    char loc_return[128], enc_return[128];

    set_perr("oxim");
    set_lc_ctype("", loc_return, 128, enc_return, 128, OXIMMSG_ERROR);
    locale->lc_ctype = (char *)strdup(loc_return);
    locale->encoding = (char *)strdup(enc_return);

    set_lc_messages("", loc_return, 128);
    locale->lc_messages = (char *)strdup(loc_return);

    split_lc_ctype(locale->lc_ctype, loc_return, enc_return);
    locale->locale = (char *)strdup(loc_return);

    if (XSupportsLocale() != True)
        perr(OXIMMSG_ERROR, 
	     N_("X locale \"%s\" is not supported by your system.\n"),
	     locale->lc_ctype);
    else
	xim_add_locale(locale->lc_ctype);
}

static void
print_usage(void)
{
    perr(OXIMMSG_EMPTY,
	N_("OXIM (Open X Input Method) version %s.\n" 
	   "(module ver: %s, syscin ver: %s).\n"),
	PACKAGE_VERSION, MODULE_VERSION, SYSCIN_VERSION);

    perr(OXIMMSG_EMPTY,
     N_("\n"
	"Usage:  oxim [-h] [-m module] [-d DISPLAY] [-x XIM_name] [-r rcfile]\n"
	"             [-U user_dir] [-D default_dir] [-v n]\n"
	"Options:  -h: print this message.\n"
	"          -m: print the comment of module \"mod_name\"\n"
	"          -d: set X Display.\n"
	"          -x: register XIM server name to Xlib.\n"
	"          -r: set user specified rcfile.\n"
	"          -U: set user oxim data directory.\n"
	"          -D: set default oxim data directory.\n"
	"          -v: set debug level to n.\n\n"));
    perr(OXIMMSG_EMPTY,
     N_("Useful environment variables:  $OXIM_RCFILE.\n\n"));
}

static void
command_switch(int argc, char **argv)
/* Command line arguements. */
{
    int  rev;
    char *rcfile=NULL, *mod_comment=NULL;
    oxim_rc_t *xrc = &(oxim_core.oxim_rc);
#ifdef HPUX
    extern char *optarg;
    extern int opterr, optopt;
#endif
/*
 *  Command line arguement & preparing for oxim_rc_t.
 */
    oxim_setlocale();
    xrc->argc = argc;
    xrc->argv = argv;

    opterr = 0;
    while ((rev = getopt(argc, argv, "hm:d:x:r:U:D:v:")) != EOF) {
        switch (rev) {
        case 'h':
            print_usage();
            exit(0);
	case 'm':
	    mod_comment = optarg;
	    break;
	case 'd':
	    strncpy(oxim_core.irc->display_name, optarg,
			sizeof(oxim_core.irc->display_name));
	    break;
	case 'x':
	    strncpy(oxim_core.irc->xim_name, optarg, 
			sizeof(oxim_core.irc->xim_name));
	    break;
        case 'r':
            rcfile = optarg;
            break;
	case 'U':
	    xrc->user_dir = optarg;
	    break;
	case 'D':
	    xrc->default_dir = optarg;
	    break;
	case 'v':
	    verbose = atoi(optarg);
	    break;
        case '?':
            perr(OXIMMSG_ERROR, N_("unknown option  -%c.\n"), optopt);
            break;
        }
    }
    check_oxim_path(xrc, OXIMMSG_ERROR);
    read_oximrc(xrc, rcfile);
/*
 *  OXIM perface.
 */
    perr(OXIMMSG_EMPTY,
	N_("OXIM (Open X Input Method) version %s.\n" 
	   "(module ver: %s, syscin ver: %s).\n"
           "(use \"-h\" option for help)\n\n"),
	PACKAGE_VERSION, MODULE_VERSION, SYSCIN_VERSION);
    perr(OXIMMSG_NORMAL, N_("locale \"%s\" encoding \"%s\"\n"),
	xrc->locale.lc_ctype, xrc->locale.encoding);
/*
 *  Print the comment of the module.
 */
    if (mod_comment) {
	mod_header_t *modp;
	if ((modp = load_module(mod_comment, MTYPE_IM, MODULE_VERSION,
		&(oxim_core.oxim_rc), NULL)) != NULL)
	    module_comment(modp, mod_comment);
	else
	    perr(OXIMMSG_EMPTY, N_("cannot open module: %s\n"), mod_comment);
	exit(0);
    }
}


/*----------------------------------------------------------------------------

        Resource Reading & Checking.

----------------------------------------------------------------------------*/

static void
read_core_config(void)
{
    oxim_rc_t *xrc = &(oxim_core.oxim_rc);
    char *cmd[1], value[256];
/*
 *  GUI config reading.
 */

    cmd[0] = "DEFAULT_FONT_NAME";
    if (get_resource(xrc, cmd, value, 256, 1))
        set_data(oxim_core.irc->default_font_name, RC_STRARR, value, 0, 
			sizeof(oxim_core.irc->default_font_name));

    cmd[0] = "DEFAULT_FONT_SIZE";
    if (get_resource(xrc, cmd, value, 256, 1))
        set_data(&oxim_core.irc->default_font_size, RC_INT, value, 0, 0);

    cmd[0] = "ROOT_FONT_SIZE";
    if (get_resource(xrc, cmd, value, 256, 1))
        set_data(&oxim_core.irc->root_font_size, RC_INT, value, 0, 0);

    cmd[0] = "PREEDIT_FONT_SIZE";
    if (get_resource(xrc, cmd, value, 256, 1))
        set_data(&oxim_core.irc->preedit_font_size, RC_INT, value, 0, 0);

    cmd[0] = "STATUS_FONT_SIZE";
    if (get_resource(xrc, cmd, value, 256, 1))
        set_data(&oxim_core.irc->status_font_size, RC_INT, value, 0, 0);

    cmd[0] = "XKILL_DISABLE";
    if (get_resource(xrc, cmd, value, 256, 1))
	set_data(&(oxim_core.oxim_mode), RC_IFLAG, value, OXIM_XKILL_OFF, 0);
    cmd[0] = "ICCHECK_DISABLE";
    if (get_resource(xrc, cmd, value, 256, 1))
	set_data(&(oxim_core.oxim_mode), RC_IFLAG, value, OXIM_ICCHECK_OFF, 0);
    cmd[0] = "SINGLE_IM_CONTEXT";
    if (get_resource(xrc, cmd, value, 256, 1))
	set_data(&(oxim_core.oxim_mode), RC_IFLAG, value, OXIM_SINGLE_IMC, 0);
    cmd[0] = "IM_FOCUS_ON";
    if (get_resource(xrc, cmd, value, 256, 1))
	set_data(&(oxim_core.oxim_mode), RC_IFLAG, value, OXIM_IM_FOCUS, 0);
    cmd[0] = "KEEP_POSITION_ON";
    if (get_resource(xrc, cmd, value, 256, 1))
	set_data(&(oxim_core.oxim_mode), RC_IFLAG, value, OXIM_KEEP_POSITION,0);
    cmd[0] = "DIFF_BEEP";
    if (get_resource(xrc, cmd, value, 256, 1))
	set_data(&(oxim_core.oxim_mode), RC_IFLAG, value, OXIM_DIFF_BEEP,0);

    cmd[0] = "FKEY_ZHEN";
    if (get_resource(xrc, cmd, value, 256, 1))
	set_funckey(FKEY_ZHEN, value);
    cmd[0] = "FKEY_2BSB";
    if (get_resource(xrc, cmd, value, 256, 1))
	set_funckey(FKEY_2BSB, value);
    cmd[0] = "FKEY_CIRIM";
    if (get_resource(xrc, cmd, value, 256, 1))
	set_funckey(FKEY_CIRIM, value);
    cmd[0] = "FKEY_CIRRIM";
    if (get_resource(xrc, cmd, value, 256, 1))
	set_funckey(FKEY_CIRRIM, value);
    cmd[0] = "FKEY_CHREP";
    if (get_resource(xrc, cmd, value, 256, 1))
	set_funckey(FKEY_CHREP, value);
    cmd[0] = "FKEY_SIMD";
    if (get_resource(xrc, cmd, value, 256, 1))
	set_funckey(FKEY_SIMD, value);
    cmd[0] = "FKEY_IMFOCUS";
    if (get_resource(xrc, cmd, value, 256, 1))
	set_funckey(FKEY_IMFOCUS, value);
    cmd[0] = "FKEY_IMN";
    if (get_resource(xrc, cmd, value, 256, 1))
	set_funckey(FKEY_IMN, value);
    cmd[0] = "FKEY_QPHRASE";
    if (get_resource(xrc, cmd, value, 256, 1))
	set_funckey(FKEY_QPHRASE, value);
    check_funckey();

    cmd[0] = "INPUT_STYLE";
    if (get_resource(xrc, cmd, value, 256, 1))
        set_data(oxim_core.irc->input_styles, RC_STRARR, value, 0, 
			sizeof(oxim_core.irc->input_styles));
    cmd[0] = "KEYBOARD_TRANSLATE";
    if (get_resource(xrc, cmd, value, 256, 1))
	set_data(&(oxim_core.oxim_mode), RC_IFLAG, value,
			OXIM_KEYBOARD_TRANS, 0);
}

void
split_lc_ctype(char *loc_buf, char *loc, char *enc)
{
    char *s, *s1;

    if ((s = strchr(loc_buf, '.'))) {
	strncpy(loc, loc_buf, (int)(s-loc_buf));
	loc[(int)(s-loc_buf)] = '\0';

	s++;
	s1 = enc;
	while (*s) {
	    *s1 = (char)tolower(*s);
	    s ++;
	    s1++;
	}
	*s1 = '\0';
    }
    else {
	strcpy(loc, loc_buf);
	enc[0] = '\0';
    }
}

static void
read_core_config_locale(void)
{
    char *cmd[2], value[256], *s, loc_buf[64], *loc_name=NULL;
    char *fmt = N_("%s:\n\tsection %s: value not specified.\n");
    oxim_rc_t *xrc = &(oxim_core.oxim_rc);
    locale_t *locale = &(oxim_core.oxim_rc.locale);

    cmd[0] = "DEFAULT_IM";
    if (! get_resource(xrc, cmd, value, 256, 1))
	perr(OXIMMSG_ERROR, fmt, oxim_core.oxim_rc.rcfile, cmd[0]);
    else
        set_data(oxim_core.irc->default_im_name, RC_STRARR, value, 0,
			sizeof(oxim_core.irc->default_im_name));

    cmd[0] = "DEFAULT_IM_MODULE";
    if (! get_resource(xrc, cmd, value, 256, 1))
	perr(OXIMMSG_ERROR, fmt, cmd[0]);
    else 
	set_data(oxim_core.irc->default_im_mod_name, RC_STRARR, value, 0,
			sizeof(oxim_core.irc->default_im_mod_name));
    if ((s = strrchr(oxim_core.irc->default_im_mod_name, '.')) && 
	!strcmp(s, ".so"))
	*s = '\0';

    cmd[0] = "DEFAULT_IM_SINMD";
    if (! get_resource(xrc, cmd, value, 256, 1))
	perr(OXIMMSG_ERROR, fmt, cmd[0]);
    else
	set_data(oxim_core.irc->default_im_sinmd_name, RC_STRARR, value, 0,
			sizeof(oxim_core.irc->default_im_sinmd_name));

    cmd[0] = "PHRASE";
    if (get_resource(xrc, cmd, value, 256, 1))
	set_data(oxim_core.irc->phrase_fn, RC_STRARR, value, 0,
			sizeof(oxim_core.irc->phrase_fn));

    cmd[0] = "CINPUT";
    if (! get_resource(xrc, cmd, value, 256, 1))
	perr(OXIMMSG_ERROR, fmt, cmd[0]);
    else
	set_data(oxim_core.irc->im_objname, RC_STRARR, value, 0,
			sizeof(oxim_core.irc->im_objname));
}

static void
read_core_config_IM(void)
{
    char *cmd[2], value[256], *s, *s1, objname[100], objenc[100];
    char *fmt = N_("%s:\n\tIM section \"%s\": %s: %s.\n");
    oxim_rc_t *xrc = &(oxim_core.oxim_rc);
    locale_t *locale = &(oxim_core.oxim_rc.locale);
    int setkey;
/*
 *  Go to each CINPUT sub-node and read important keywords.
 */
    s1 = oxim_core.irc->im_objname;
    while (get_word(&s1, objname, 100, NULL)) {
	cmd[0] = objname;
	cmd[1] = "SETKEY";
        if (! get_resource(xrc, cmd, value, 256, 2))
	    perr(OXIMMSG_ERROR, fmt, oxim_core.oxim_rc.rcfile, objname, cmd[1], N_("value not specified"));

	/* setkey should be uniquely defined */
	set_data(&setkey, RC_INT, value, 0, 0);
	if (setkey < 0 || setkey > MAX_IM_ENTRY)
	    perr(OXIMMSG_ERROR,N_(fmt),oxim_core.oxim_rc.rcfile,objname,cmd[1],
		N_("invalid range of value"));
	else if(IM_check_registered(setkey) == True)
	    perr(OXIMMSG_ERROR,N_(fmt),oxim_core.oxim_rc.rcfile,objname,cmd[1],
		N_("value conflict with other module"));

	cmd[1] = "MODULE";
        if (get_resource(xrc, cmd, value, 256, 2)) {
            if ((s = strrchr(value, '.')) && !strcmp(s, ".so"))
		*s = '\0';
	    IM_register(setkey, value, cmd[0], locale->locale);
	}
	else
	    IM_register(setkey, oxim_core.irc->default_im_mod_name, 
		       cmd[0], locale->locale);
    }
}


/*----------------------------------------------------------------------------

        Initialization

----------------------------------------------------------------------------*/
static void load_support_locales(void)
{
    locale_t *locale = &(oxim_core.oxim_rc.locale);

    unsigned int flen = strlen(OXIM_DEFAULT_RCDIR) + strlen(OXIM_DEFAULT_LOCALE_LIST) + 2;
    char *locale_list = oxim_malloc(flen, True);
    char loc_buf[66], loc_name[32], loc_enc[32];

    sprintf(locale_list, "%s/%s", OXIM_DEFAULT_RCDIR, OXIM_DEFAULT_LOCALE_LIST);

    if (check_file_exist(locale_list, FTYPE_FILE))
    {
	FILE *fp = fopen(locale_list, "r");
	while (fgets(loc_buf, 66, fp))
	{
	    char *p = loc_buf;
	    unsigned int len = strlen(p);
	    while (len && p[len-1] <= ' ')
	    {
		p[--len] = '\0';
	    }
	    while (len && *p <= ' ')
	    {
		p++;
		len --;
	    }

	    if (len && *p != '#')
	    {
		xim_add_locale(p);
	    }
	}
	fclose(fp);
    }
    else
    {
	perr(OXIMMSG_ERROR, N_("File: '%s' Not exist.\n"), locale_list);
    }
    free(locale_list);
}

void load_syscin(void)
{
    gzFile *fp=NULL;
    int len;
    oxim_rc_t *xrc = &(oxim_core.oxim_rc);
    char buf[40], truefn[256], *sub_path = "tables/";
    char inpn_english[CIN_CNAME_LENGTH];
    char inpn_sbyte[CIN_CNAME_LENGTH];
    char inpn_2bytes[CIN_CNAME_LENGTH];
    uch_t ascii[N_ASCII_KEY];
    charcode_t ccp[UCH_SIZE];

    if (check_datafile("sys.tab", sub_path, xrc, truefn, 256) == True)
	fp = gzopen(truefn, "rb");
    else
	perr(OXIMMSG_ERROR, N_("sys.tab does not exist.\n"));
    if (gzread(fp, buf, MODULE_ID_SIZE) != MODULE_ID_SIZE ||
	strcmp(buf, "syscin"))
	perr(OXIMMSG_ERROR, N_("invalid tab file: %s.\n"), truefn);
    len = sizeof(SYSCIN_VERSION);
    if (gzread(fp, buf, len) != len || strcmp(SYSCIN_VERSION, buf)!=0)
	perr(OXIMMSG_ERROR, N_("invalid sys.tab version: %s.\n"), buf);

    if (gzread(fp, inpn_english, CIN_CNAME_LENGTH)
		!= CIN_CNAME_LENGTH ||
	gzread(fp, inpn_sbyte, CIN_CNAME_LENGTH)
		!= CIN_CNAME_LENGTH ||
	gzread(fp, inpn_2bytes, CIN_CNAME_LENGTH)
		!= CIN_CNAME_LENGTH ||
	gzread(fp, ascii, sizeof(uch_t) * N_ASCII_KEY) != (sizeof(uch_t) * N_ASCII_KEY) ||
	gzread(fp, ccp, sizeof(charcode_t) * UCH_SIZE) != (sizeof(charcode_t) * UCH_SIZE))
	perr(OXIMMSG_ERROR, N_("sys.tab reading error.\n"));
    gzclose(fp);

    oxim_core.gui.inpn_english = (char *)strdup(inpn_english);
    oxim_core.gui.inpn_sbyte   = (char *)strdup(inpn_sbyte);
    oxim_core.gui.inpn_2bytes  = (char *)strdup(inpn_2bytes);
    fullascii_init(ascii);
    ccode_init(ccp, UCH_SIZE);
}

static void
load_default_IM(void)
{
    int idx;
    imodule_t *imodp=NULL;
/*
 *  Try to load the default input method.
 */
    imodp = IM_search(oxim_core.irc->default_im_name, 
	    oxim_core.oxim_rc.locale.locale, &idx, &(oxim_core.oxim_rc));
    if (imodp) {
	oxim_core.im_focus = oxim_core.default_im = (inp_state_t)idx;
	return;
    }
/*
 *  If false, try to load any specified input method.
 */
    imodp = IM_get_next(0, &idx, &(oxim_core.oxim_rc));
    if (imodp) {
	strncpy(oxim_core.irc->default_im_name, imodp->objname,
		sizeof(oxim_core.irc->default_im_name));
	oxim_core.im_focus = oxim_core.default_im = (inp_state_t)idx;
    }
    else
	perr(OXIMMSG_ERROR, N_("no valid input method loaded.\n"));
}
    
static void
load_default_sinmd(void)
{
    int idx;
    imodule_t *imodp=NULL;
/*
 *  Check the default im_sinmd.
 */
    if (! strcmp(oxim_core.irc->default_im_sinmd_name, "DEFAULT"))
	oxim_core.oxim_mode |= OXIM_SHOW_CINPUT;
    else {
	imodp = IM_search(oxim_core.irc->default_im_sinmd_name, 
	    oxim_core.oxim_rc.locale.locale, &idx, &(oxim_core.oxim_rc));
	if (imodp)
	    oxim_core.default_im_sinmd = (inp_state_t)idx;
	else
	    oxim_core.oxim_mode |= OXIM_SHOW_CINPUT;
    }
}

/*----------------------------------------------------------------------------

        Main Program.

----------------------------------------------------------------------------*/

void xim_terminate(void);

void sighandler(int sig)
{
    if (sig == SIGQUIT) {
	DebugLog(1, ("catch signal: SIGQUIT\n"));
    }
    else if (sig == SIGTERM) {
	DebugLog(1, ("catch signal: SIGTERM\n"));
    }
    else {
	DebugLog(1, ("catch signal: SIGINT\n"));
    }
    if (sig == SIGQUIT)
	return;

    if (oxim_core.ic != NULL && (oxim_core.ic->ic_state & IC_FOCUS))
	oxim_core.oxim_mode |= OXIM_RUN_KILL;
    else {
	xim_terminate();
	IM_free_all();
	exit(0);
    }
}

int
main(int argc, char **argv)
{
    signal(SIGQUIT, sighandler);
    signal(SIGTERM, sighandler);
    signal(SIGINT,  sighandler);

    oxim_core.irc = oxim_malloc(sizeof(inner_rc_t), 1);
/*
 *  Option piroity:  command line > environment variable > user rcfile.
 */
    command_switch(argc, argv);
    read_core_config();
    read_core_config_locale();
    read_core_config_IM();

    load_support_locales();

    load_syscin();
    load_default_IM();
    load_default_sinmd();
    qphrase_init(&(oxim_core.oxim_rc), oxim_core.irc->phrase_fn);
    gui_init(&oxim_core);
    xim_init(&oxim_core);
    free(oxim_core.irc);

    gui_loop(&oxim_core);
    return 0;
}
